#!/usr/bin/perl
##
## Convert photos to show on Philips Photo Frame 9FF2
## Version 1.0
## Written by Joakim Andersson, joakim at iqcc dot se
##
## Syntax:
## perl convert_to_photoframe.pl <dir> <dir>...
## or
## perl convert_to_photoframe.pl all
##
## Each dir will be written to a corresponding ALBUMxxx
## which you simply copies to your memorycard
## What album that corresponds to a dir will be maintained
## in a textfile album_info.txt together with the album-name
## and a ID that the photoframe needs somehow (probably
## slideshow-settings or so).
##
## If 'all', the remembered mapping of dir to album is ignored
## and every dir will be done. Useful if you want to reorder
## the created albums. It's best to delete old albums first
##
## Required modules can be found at:
## http://search.cpan.org/~jcristy/PerlMagick-6.32/Magick.pm
## http://search.cpan.org/~bettelli/Image-MetaData-JPEG-0.15/lib/Image/MetaData/JPEG.pod
##

use strict;
use Image::Magick;
use Image::MetaData::JPEG;

################################################################
# Settings - The defaults should be fine for most people

# Scale images down to this size, maintaining aspect
# Images will not be scaled up if smaller, but will be centered
# on a black canvas of this size (to fit the text)
# Note that the resolution of the lcd is 640x480, but Philips
# own program scales to this size, and slideshow effects
# looks better at this resolution than 640x480
my $WIDTH = 720;
my $HEIGHT = 540;

# Despite it's name, UnsharpMask adds sharpness to the picture
# When scaled down, the picture easily looses sharpness
# Performing this really makes a difference on the photoframe
my $UNSHARP_ENABLED = 1;

# Jpeg Quality when saving the photos, max 100
# Adjust after your need, but beware of artifacts
# 97 gives me ~1200 photos/256MB
# 95 cuts sizes with 50-66% or so
my $JPEG_QUALITY = 97;

# If and how information about the photo should be added
# Directory-name, photo-number and time (read from EXIF)
# Indent:
# The text stamped on the photo shall be this far in from
# the edges. The default indents are adjusted to 720x540
# Font and size:
# The font must be on your system. The default one was on my
# by default (Fedora Core 3) and looks excellent on the display.
# Not all fonts in system seems to work, I dunno.
my $TEXT_ENABLED = 1;
my $TEXT_LEFT_INDENT = 7;
my $TEXT_RIGHT_INDENT = 13;
my $TEXT_BOTTOM_INDENT = 45;
my $TEXT_FONT = 'NewCenturySchlbk-BoldItalic';
my $TEXT_SIZE = 12; 

# Controls how ALBUMxxx is created
# The firmware in my frame doesn't allow me to easily scroll to
# the album with the highest number, so I prefer to create newer
# albums with lower numbers than the current ones
# Use a positive step to go upwards
my $DIR_START = 400;
my $DIR_STEPS = -5;

# End of settings
###############################################################

unless ($ARGV[0])
{
	print <<END;
Syntax:
perl $0 <dir> <dir>...
 or
perl $0 all <basedir>
END
	exit;
}

open (ID, "curr_id.dat");
my $curr_id = <ID>;
close ID;

# If we have no id we start at a decent point
$curr_id ||= 1400;

##
## Reading current albums in the dir
##

my $lowest_dir = 0;
my $highest_dir = 0;
my %current_dirs = ();

opendir (OUTDIR, '.');
foreach (readdir(OUTDIR))
{
	next unless (/^(\d\d\d)ALBUM$/);
	my $num = $1+0;
	
	$current_dirs{$num} = 1;
	$lowest_dir = $num if ($lowest_dir && $num < $lowest_dir);
	$highest_dir = $num if ($highest_dir && $num > $highest_dir);
}

$DIR_STEPS ||= 1;

my $curr_dir = $highest_dir;
$curr_dir = $lowest_dir if ($DIR_STEPS < 0);
$curr_dir ||= $DIR_START;

##
## If you give the sole argument 'all'
## the whole album-structure is re-created
##

my @dirs = @ARGV;
my $ignore_curr_albumnum = 0;

if ($ARGV[0] eq 'all')
{
	@dirs = ();
	$ignore_curr_albumnum = 1;
	%current_dirs = ();
	$curr_dir = $DIR_START;
	
	my $base_dir = $ARGV[1] || '.';
	
	opendir (DIR, $base_dir) || die "Unable to open base directory '$base_dir'!";
	foreach (sort readdir(DIR))
	{
		next if (/^\d\d\dALBUM$/);
		next if (/^\./);
		next unless (-d $_);
		
		push(@dirs, $_);
	}
	closedir DIR;
}

##
## Let's rock!
##

my $tot_insize = 0;
my $tot_outsize = 0;
my $tot_count = 0;

foreach my $dir (@dirs)
{
	next unless (-d $dir);
	next if (/^\./);
	next if ($dir =~ /^(\d\d\d)ALBUM$/);
	
	##
	## Counts the photos in the dir, for showing in the text
	##
	
	my $dir_count = 0;
	my @files = ();
	
	opendir(DIR, $dir) || print "Unable to open directory '$dir'!";
	foreach my $file (sort readdir(DIR))
	{
		next unless ($file =~ /\.(jpg|jpeg|gif)$/i);
		$dir_count++;
		push (@files, $file);
	}
	closedir DIR;
	
	unless ($dir_count)
	{
		print "Skips this dir, no photos\n\n";
		next;
	}
	
	##
	## Reads info about this frame
	##
	
	my $infofile = "$dir/album_info.txt";
	my %INFO = ();
	open (INFO, $infofile);
	while (<INFO>)
	{
		next if (/^#/);
		tr/\n\r//d;
		next unless ($_);
		
		my ($name, $value) = split(/\s*=\s*/, $_, 2);
		$INFO{$name} = $value;
	}
	close INFO;
	
	# Removes current album-info if 'all' is given
	delete $INFO{'OUT'} if ($ignore_curr_albumnum);
	
	# Name of output dir
	unless ($INFO{'OUT'})
	{
		##
		## Select new dir
		##
		
		while ($current_dirs{$curr_dir})
		{
			$curr_dir += $DIR_STEPS;
		}
		
		my $pad = '0' x (3 - length($curr_dir));
		$INFO{'OUT'} = "../${pad}${curr_dir}ALBUM";
		
		$current_dirs{$curr_dir} = 1;
	}
	
	# Name i LONGNAME, name of album shown in photoframe
	$INFO{'NAME'} ||= $dir;
	$INFO{'NAME'} =~ s/^(\d\d\d\d)(\-\d\d)?(\-\d\d)?/$1/;	# Strips month and day
	$INFO{'NAME'} =~ tr/-/aaoAAO_/;
	$INFO{'NAME'} =~ tr/A-Za-z0-9_/A-Za-z0-9_/d;
	$INFO{'NAME'} = substr($INFO{'NAME'}, 0, 24);
	
	# ID, used somehow by photoframe
	unless ($INFO{'ID'})
	{
		# Picks next id, unique hopefully
		$INFO{'ID'} = $curr_id++;

		open (ID, ">curr_id.dat") || die "Unable to write to curr_id.dat!";
		print ID $curr_id;
		close ID;
	}
	
	# Saves info for next run
	open (INFO, ">$infofile");
	foreach (qw(OUT NAME ID))
	{
		print INFO "$_\t= $INFO{$_}\r\n";
	}
	close INFO;
	
	##
	## Creates the album-dir
	##
	
	my $outdir = "$dir/$INFO{'OUT'}";
	mkdir $outdir unless (-e $outdir);
	
	print "Converting $dir -> $INFO{'OUT'} ($INFO{'NAME'}, ID $INFO{'ID'})\n";
	
	##
	## Sets the name by creating the LONGNAME-file
	## Also, writes the info in cleartext for easy identification
	##
	
	# Drakbat 1998            1332
	my $longinfo = $INFO{'NAME'};
	$longinfo .= ' '  x (24 - length($longinfo));
	$longinfo .= $INFO{'ID'};
	
	open (ALBUM, ">$outdir/LONGNAME");
	print ALBUM $longinfo;
	close ALBUM;
	
	open (INFO, ">$outdir/$INFO{'NAME'}");
	close INFO;
	
	##
	## Okay, converts each photo
	##
	
	my $i = 0;
	my $dir_insize = 0;
	my $dir_outsize = 0;
	
	foreach my $file (@files)
	{
		my $infile = "$dir/$file";
		my $insize = int((-s $infile) / 1024);
		
		$i++;
		my $num = '0' x (4 - length($i)) . $i;
		
		my $outname = "Phi_$num.JPG";
		my $outfile = "$outdir/$outname";
		unlink $outfile if (-e $outfile);
		
		##
		## EXIF-info
		##
		
		my $exif = new Image::MetaData::JPEG($infile);
		my $exif_ref = $exif->get_Exif_data('IMAGE_DATA', 'TEXTUAL');
		
		# D   132  DateTime                    A 20  RRRR OOOO   YYYY:MM:DD HH:MM:SS
		# A   112  Orientation                 S 1   RRRR OOOO   1-8
		
		my $datetime = $exif_ref->{'DateTime'}[0];
		my $orientation = $exif_ref->{'Orientation'}[0];
		#print Dumper($exif_ref);
		
		##
		## Opens the image with Image::Magick
		##
		
		my $image = Image::Magick->new;
		my $x = $image->Read($infile);
		warn "  $file $x\n" if "$x";
		
		##
		## Rotates the piture if the EXIF says so
		##
		
		if ($orientation)
		{
			if ($orientation == 8)
			{
				$x = $image->Rotate(degrees => 270, color => 'black');
				warn "  $file $x\n" if "$x";
				
				$orientation = 1;
			}
			elsif ($orientation == 6)
			{
				$x = $image->Rotate(degrees => 90, color => 'black');
				warn "  $file $x\n" if "$x";
				
				$orientation = 1;
			}
			elsif ($orientation != 1)
			{
				warn "  $file Unknown orientation '$orientation'!\n";
				exit;
			}
		}
		
		##
		## Calculates new width and height
		##
		
		my ($width, $height) = $image->Get('width','height');
		my $new_width  = $width;
		my $new_height = $height;
		
		if ($new_width > $WIDTH)
		{
			# Too wide, reduce height and width equally
			$new_height = int($new_height * $WIDTH / $new_width);
			$new_width = $WIDTH;
		}
		
		if ($new_height > $HEIGHT)
		{
			# Too high, reduce height and width equally
			$new_width = int($new_width * $HEIGHT / $new_height);
			$new_height = $HEIGHT;
		}
		
		##
		## Scales the picure
		##
		
		$x = $image->Resize(height => $new_height, width => $new_width, filter => 'Lanczos');
		warn "  $file $x\n" if "$x";
		
		##
		## Smaller photos are centered on a black canvas
		##
		
		if ($new_height < $HEIGHT || $new_width < $WIDTH)
		{
			my $image2 = Image::Magick->new;
			$image2->Set(size=>"${WIDTH}x${HEIGHT}");
			$image2->ReadImage('xc:black');
			
			$x = $image2->Composite(image => $image, compose => 'Over', gravity => 'Center');
			warn "  $file $x\n" if "$x";
			
			$image = $image2;
		}
		
		##
		## Enhances sharpness a tiny bit
		##
		
		if ($UNSHARP_ENABLED)
		{
			$x = $image->UnsharpMask(radius => 0.5, sigma => 0.5, amount => 1.0, threshold => 0.02);
			warn "  $file $x\n" if "$x";
		}
		
		##
		## Add the text, if wanted
		##
		
		if ($TEXT_ENABLED)
		{
			my $y = $HEIGHT - $TEXT_BOTTOM_INDENT;
			
			##
			## Catalog-name to the left
			##
			
			my $show_dir = $dir;
			$dir =~ s/\/$//;
			$show_dir = $1 if ($dir =~ /[^\/](.+)$/);
			
			$x = $image->Annotate(text => $dir, pointsize => $TEXT_SIZE, x => $TEXT_LEFT_INDENT, y => $y, align => 'left', font => $TEXT_FONT, fill => 'black', antialias => 'true', strokewidth => 2, stroke => 'black');
			warn "  $file $x\n" if "$x";
			
			$x = $image->Annotate(text => $dir, pointsize => $TEXT_SIZE, x => $TEXT_LEFT_INDENT, y => $y, align => 'left', font => $TEXT_FONT, fill => 'white', antialias => 'true');
			warn "  $file $x\n" if "$x";
			
			##
			## Current/total count 
			##
			
			$x = $image->Annotate(text => "$i/$dir_count", pointsize => $TEXT_SIZE, x => int($WIDTH/2), y => $y, align => 'center', font => $TEXT_FONT, fill => 'black', antialias => 'true', strokewidth => 2, stroke => 'black');
			warn "  $file $x\n" if "$x";
			
			$x = $image->Annotate(text => "$i/$dir_count", pointsize => $TEXT_SIZE, x => int($WIDTH/2), y => $y, align => 'center', font => $TEXT_FONT, fill => 'white', antialias => 'true');
			warn "  $file $x\n" if "$x";
			
			##
			## Datetime, if present in the EXIF
			##
			
			if ($datetime)
			{
				my $time = $datetime;
				$time = "$1-$2-$3 $4:$5" if ($time =~ /^(\d\d\d\d).(\d\d).(\d\d).(\d\d).(\d\d)/);
				
				$x = $image->Annotate(text => $time, pointsize => $TEXT_SIZE, x => $WIDTH-$TEXT_RIGHT_INDENT, y => $y, align => 'right', font => $TEXT_FONT, fill => 'black', antialias => 'true', strokewidth => 2, stroke => 'black');
				warn "  $file $x\n" if "$x";
				
				$x = $image->Annotate(text => $time, pointsize => $TEXT_SIZE, x => $WIDTH-$TEXT_RIGHT_INDENT, y => $y, align => 'right', font => $TEXT_FONT, fill => 'white', antialias => 'true');
				warn "  $file $x\n" if "$x";
			}
		}
		
		##
		## Saves as jpeg with wanted quality
		##
		
		$x = $image->Write(filename => $outfile, compression => 'JPEG', quality => $JPEG_QUALITY);
		warn "  $file $x\n" if "$x";
		
		##
		## Image::Magick maintains EXIF-orientation
		## Since we're rotated the photo we reset the orientation
		##
		
		my $exif_file = new Image::MetaData::JPEG($outfile);
		my $exif_ref = $exif_file->provide_app1_Exif_segment();
		
		if ($datetime)
		{
			$x = $exif_file->set_Exif_data({ 'DateTime' => $datetime }, 'IMAGE_DATA', 'REPLACE');
			print "  $file Kunde inte stta EXIF DateTime = $datetime fr $outname!\n" if (%$x);
		}
		
		if ($orientation)
		{	
			$x = $exif_file->set_Exif_data({ 'Orientation' => $orientation }, 'IMAGE_DATA', 'REPLACE');
			print "Kunde inte stta EXIF Orientation = $orientation fr $outname!\n" if (%$x);
		}
		
		$exif_file->save($outfile);
		
		##
		## Photo done!
		##
		
		my $outsize = int((-s $outfile) / 1024);
		my $ratio = int( $insize / $outsize );
		
		print "  $file (${width}x$height, $insize kB) -> $outname (${new_width}x$new_height, $outsize kB), $ratio gnger mindre\n";
		
		$dir_insize += $insize;
		$dir_outsize += $outsize;
	}
	
	##
	## Cleans old photos left over in the album
	##
	
	while (1)
	{
		$i++;
		my $num = '0' x (4 - length($i)) . $i;
		
		my $outname = "Phi_$num.JPG";
		my $outfile = "$outdir/$outname";
		last unless (-e $outfile);
		
		unlink $outfile;
		print "  Raderat verbliven bild $outname\n";
	}
	
	##
	## The dir is completed!
	##
	
	$tot_insize  += $dir_insize;
	$tot_outsize += $dir_outsize;
	$tot_count   += $dir_count;

	my $ratio = 'NA';
	if ($dir_outsize > 0)
	{
		$ratio = int($dir_insize/$dir_outsize);
	}
	
	print "Dir done, $dir_count photos processed, $dir_outsize kB (was $dir_insize kB => ratio $ratio), total $tot_outsize kB this far\n\n";
}

##
## Alles klar!
##

my $ratio = 'NA';
if ($tot_outsize > 0)
{
	$ratio = int($tot_insize/$tot_outsize);
}

print "All dirs klara, $tot_count photos processed, $tot_outsize kB (was $tot_insize kB => ratio $ratio)!\n";

exit;

